;;;; simple-units.scm
;;;; Kon Lovett, Aug '10

;; Issues
;;
;; - Quantity is not 1st-class.
;;
;; This requires extending the numeric tower!
;;
;; - No "dimension".
;;
;; - No units-systems catalog.
;;
;; - Cannot remove a units from a units-system.
;;
;; - Cannot extend a units.
;;
;; - The 'units-scale' variable is a kludge.
;;
;; - Cannot create derived units.
;;
;; (define mile/h (make-derived-unit mile / h))
;; (define 23mile/h (make-quantity 23 mile/h))

;; Notes
;;
;; - Rather than eager value resolution could be lazy.
;;
;;  - When a value is not a factor use a promise & then force at point of
;;  reference.

(module simple-units

  (;export
    ;
    units-scale
    units-number?
    units-number-system
    ;
    (make-units *make-units)
    (make-units* *make-units)
    units? check-units error-units
    units-name
    units-base-name
    units-units-system
    units-unit-names
    units-unit
    ;
    unit? check-unit error-unit
    unit-name
    unit-factor
    ;
    (make-quantity *make-quantity)
    ;
    default-units-system
    make-units-system
    units-system? check-units-system error-units-system
    units-system-name
    units-system-catalog
    units-system-units
    units-system-units/unit)

  (import
    scheme
    chicken
    (only srfi-1 every find delete)
    (only format-compiler-base format)
    (only record-variants define-record-type-variant)
    (only list-utils plist->alist)
    (only type-checks define-check+error-type)
    (only type-errors warning-argument-type))

  (require-library
    srfi-1
    format-compiler-base
    record-variants
    list-utils
    type-checks
    type-errors)

;;

(define-syntax define-variable
  (syntax-rules ()
    ((_ name init pred typ)
      (define name
        (let ((cache init)
              (_pred pred)
              (_typ typ) )
          (lambda args
            (if (null? args) cache
              (let ((arg (car args)))
                (if (_pred arg) (set! cache arg)
                  (warning-argument-type 'name arg _typ) ) ) ) ) ) ) ) ) )

(define-variable units-scale * procedure? 'procedure)

(define-variable units-number? number? procedure? 'procedure)

(define (units-number-system scale numberp)
  (units-scale scale)
  (units-number? numberp) )

;; Basic Types

; exposition purposes only

(define unit-name? symbol?)
(define units-name? symbol?)
(define units-system-name? symbol?)

(define-check+error-type unit-name)
(define-check+error-type units-name)
(define-check+error-type units-system-name)

(define (factor? x) ((units-number?) x))

#; ;UNUSED
(define-check+error-type factor)

;; Unit Type

(define (%make-unit name factor) (cons name factor))
(define (%unit-name unit) (car unit))
(define (%unit-factor unit) (cdr unit))

(define (%unit-ref name units-tbl) (assq name units-tbl))

(define (unit? obj)
  (and (pair? obj)
       (unit-name? (car obj))
       (factor? (cdr obj))) )

(define-check+error-type unit)

(define (unit-name unit)
  (%unit-name (check-unit 'unit-name  unit)) )

(define (unit-factor unit)
  (%unit-factor (check-unit 'unit-factor  unit)) )

;; Unit-Set Type

(define-record-type-variant units (unsafe unchecked inline)
  (%make-units nam bas tbl qdb)
  %units?
  (nam %units-name)
  (bas %units-base)
  (tbl %units-table)
  (qdb %units-database) )

; returns the actual unit or #f
; do not expose!
(define (%units-unit units name)
  (%unit-ref name (%units-table units)) )

(define-check+error-type units %units?)

(define (units? obj) (%units? obj))

(define (units-name units)
  (%units-name (check-units 'units-name units)) )

(define (units-base-name units)
  (%unit-name (%units-base (check-units 'units-base-unit-nam units))) )

(define (units-unit-names units)
  (map car (%units-table (check-units 'units-list units))) )

(define (units-unit units name)
  (let ((unit (%units-unit (check-units 'units-list units) (check-unit-name 'units-unit name))))
    (%make-unit (%unit-name unit) (%unit-factor unit)) ) )

(define (units-units-system units)
  (%units-database (check-units 'units-list units)) )

;; Units-System Type

(define (make-units-system name)
  (cons (check-units-system-name 'make-units-system name) '()) )

(define (%units-system-units db) (cdr db))

(define (units-system? obj)
  (and (pair? obj)
       (units-system-name? (car obj))
       (every units? (cdr obj))) )

(define-check+error-type units-system)

(define-variable default-units-system (make-units-system 'default)
  units-system? 'units-system)

(define (%add-units! units #!optional (db (default-units-system)))
  (unless (memq units (%units-system-units db))
    (set-cdr! db (cons units (%units-system-units db))) ) )

#; ;Too many implications
(define (%remove-units! units #!optional (db (default-units-system)))
  (set-cdr! db (delete units (%units-system-units db) eq?)) )

(define (%find-units name #!optional (db (default-units-system)))
  (find (lambda (x) (eq? name (%units-name x))) (%units-system-units db)) )

(define (%find-units/unit name #!optional (db (default-units-system)))
  (find (lambda (units) (and (%units-unit units name) units)) (%units-system-units db)) )

; returns the actual unit or #f
; do not expose!
(define (%find-unit name #!optional (db (default-units-system)))
  (find (lambda (units) (%units-unit units name)) (%units-system-units db)) )

;;

(define (units-system-name #!optional (db (default-units-system)))
  (car (check-units-system 'units-system-units db)) )

(define (units-system-catalog #!optional (db (default-units-system)))
  (cdr (check-units-system 'units-system-catalog db)) )

; returns the units in unit-system by name
(define (units-system-units name #!optional (db (default-units-system)))
  (%find-units name (check-units-system 'units-system-units db)) )

; returns the units in unit-system with unit by name
(define (units-system-units/unit name #!optional (db (default-units-system)))
  (%find-units/unit name (check-units-system 'units-system-units db)) )

;; Quantity Value

(define-syntax make-quantity
  (syntax-rules (system:)
    ((_ ?n ?name)
      (make-quantity ?n ?name (units-system-units/unit '?name)) )
    ((_ ?n ?name system: ?db)
      (make-quantity ?n ?name (units-system-units/unit '?name ?db)) )
    ((_ ?n ?name ?units-name system: ?db)
      (make-quantity ?n ?name (units-system-units '?units-name ?db)) )
    ((_ ?n ?name ?units)
      (*make-quantity ?n '?name ?units) ) ) )

(define (*make-quantity n name units)
  (check-units 'make-quantity units)
  ((units-scale) n (%unit-factor (%units-unit units name))) )

;;

(define-syntax quotify
  (syntax-rules ()
    ((_)
      '() )
    ((_ (?item ...))
      (list '?item ...) )
    ((_ ?item)
      '?item )
    ((_ ?item ...)
      (list (quotify ?item) ...) ) ) )

(define-syntax make-units*
  (syntax-rules (system:)
    ((_ ?name ?base-name system: ?db ?units ...)
      (*make-units '?name '?base-name (quotify ?units ...) #t ?db) )
    ((_ ?name ?base-name ?units ...)
      (*make-units '?name '?base-name (quotify ?units ...) #t) ) ) )

(define-syntax make-units
  (syntax-rules (system:)
    ((_ ?name ?base-name system: ?db ?units ...)
      (*make-units '?name '?base-name (quotify ?units ...) #f ?db) )
    ((_ ?name ?base-name ?units ...)
      (*make-units '?name '?base-name (quotify ?units ...) #f) ) ) )

; Any external unit dependencies must be on resolved units!
; Only internal unresolved unit dependencies handled.

(define (*make-units name base-name units #!optional (anonymous? #f) (db (default-units-system)))

  ; cache the operator
  (define scale (units-scale))

  ; same structure but different interpretation
  (define %unit-value %unit-factor)

  ; <unit-spec> ::= (<symbol> . <value-spec-item>...)
  ; <value-spec-item> ::= <symbol>
  ;                     | <number>
  ;                     | (<value-spec-item>...) ;what does this mean?

  (define (check-unit unit-spec)

    ; check unit-spec-value
    (define (check-unit-value x)
      (cond
        ((or (factor? x) (unit-name? x))
          x )
        ((pair? x)
          (let ((hd (check-unit-value (car x))))
            ; don't follow empty tail
            (if (null? (cdr x)) hd
              (let ((tl (check-unit-value (cdr x))) )
                (if (list? hd) (append hd tl) ;flatten for now
                  (cons hd tl) ) ) ) ) )
        (else
          (error 'make-units
            "invalid unit-value - not a unit value specification" x) ) ) )

    (cons
      (check-unit-name 'make-units (car unit-spec))
      (check-unit-value (cdr unit-spec))) )

  ; <resolved-unit-spec> ::= (<symbol> . <number>)
  ;
  ; resolve-unit-spec-value = (scale (resolve-factor <value-spec-item>)...)
  ; resolve-factor =
  ;   with item = <value-spec-item>
  ;     ((factor? item) -> item)
  ;     ((name? item) -> (unit-value (resolve-unit-ref item)))
  ;
  ; resolve-unit-ref =

  (define (resolve-unit unit units trail)

    (define (error-depends name trail)
      (error 'make-units (format #f "~{ ~a~^ depends on~}" (cons name trail))) )

    (define (check-depends name trail)
      (if (memq name trail) (error-depends name trail)
          name ) )

    ; resolves the value-spec only so returns a number
    (define (resolve-value value trail)
      ; value is an unresolved unit value specificaton
      (cond
        ; already resolved?
        ((factor? value)
          value )
        ; unresolved unit reference?
        ((unit-name? value)
          ; look in this units first for a dependency
          (let* ((dep-name (check-depends value trail)) ;check for a dependency loop
                 (dep-unit (%unit-ref dep-name units)) )
            ; a dependency on unit in this units is resolved recursively
            (if dep-unit (resolve-value (%unit-value dep-unit) (cons (%unit-name dep-unit) trail))
              ; maybe dependency on unit in an external units
              (let ((dep-unit (%find-unit dep-name db)))
                (if (not dep-unit) (error 'make-units "dependence on undefined unit" dep-name)
                  (let ((val (%unit-value dep-unit)))
                    (if (factor? val) val
                      ;this shouldn't happen,
                      ; since externals must be already resolved.
                      (error 'make-units "dependence on unresolved unit" dep-name) ) ) ) ) ) ) )
        ; scale by remaining unresolved unit-value-spec
        ; we know that the 1st value-spec-item here is a number
        (else
          (scale (car value) (resolve-value (cdr value) trail)) ) ) )

    (let ((value (cdr unit)))
      ; when already resolved (w/o dependenies) just return the identity
      (if (factor? value) unit
        ; value = (factor ... unit-name)
        (let ((name (car unit)))
          ; resolve unit-value expression
          (%make-unit
            (check-depends name trail)  ;check for a dependency loop
            (resolve-value value (cons name trail))) ) ) ) )

  ;
  (check-units-name 'make-units name "units-name")
  (check-unit-name 'make-units base-name "base-unit-name")
  (check-units-system 'make-units db "units-system")
  ; Refine unitset form -> units type
  (let ((units (plist->alist units))
        (base-unit (%make-unit base-name 1)) )
    (let ((units (cons base-unit (map check-unit units))))
      (let ((units (map (cut resolve-unit <> units '()) units)))
        (let ((units (%make-units name base-unit units db)))
          (unless anonymous?
            (%add-units! units db) )
          units ) ) ) ) )

) ;module simple-units
